// GSMSMS Library
// Copyright (c) 2020 by Haute Solutions / Ed Van Every (www.haute-solutions.com)
//
// FUNCTION: This Class Library is intended to simplify command processing for SMS Enabled Devices.
//           It inherits objects and methods from the Haute Solutions GSMSMS and Haute Solutions OLED
//           class libraries.  OLED Functionality is optional (Init with NULL)
//
// HARDWARE: Please refer to the GSMSMS and OLED Class Libraries for Hardware Support Information
//
// SOFTWARE: Requires the Haute Solutions GSMSMS and Haute Solutions OLED Class Libraries
//
// LICENSE:  This library is free software: you can redistribute it and/or modify
//           it under the terms of the GNU General Public License as published by
//           the Free Software Foundation, either version 3 of the License, or
//           (at your option) any later version.
//
//           This program is distributed in the hope that it will be useful,
//           but WITHOUT ANY WARRANTY; without even the implied warranty of
//           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//           GNU General Public License for more details.
//           <http://www.gnu.org/licenses/>.
//
//
// VERSION:  SMSDEV_LIBVER "1.3"

#include "Arduino.h"
#include "SMSDEV.h"

CMDINFO GenCmds[] = {
  {"PWD:",         1, 1},
  {"DEVNUM:",      2, 1},
  {"DEVNAME:",     3, 1},
  {"HBTIME:",      4, 1},
  {"BOOTNOTIFY:",  5, 1},
  {"POLLSMS:",     6, 1},
  {"CFGINFO",      7, 1},
  {"AUTHLIST",     8, 1},
  {"RESET",        9, 1},
  {"REBOOT",      10, 1},
  {"SIMINFO",     11, 0},
  {"HWINFO",      12, 0},
  {"CELLINFO",    13, 0},
  {"OPERATORS",   14, 0},
  {"ATCMD:",      15, 0}
};

int SMSDEV::Setup(char *company, char* device, char* version, char* sig, char* cmdprefix, CMDINFO* cmds, int numcmds) {
char simcom[10];

#if defined(ESP32)
    EEPROM.begin(800);
#endif
	
	Company = company;
	Device = device;
	Version = version;
	Signature = sig;
	CmdPrefix = cmdprefix;
	SpecCmds = cmds;
	NumSpecCmds = numcmds;
	ProcessPrivs = 1;
	
	NumGenCmds = sizeof(GenCmds)/sizeof(GenCmds[0]);
	
	OLED::Init();
	OLED::Clear();
	  
  SMSDEV::DisplayWithDelay(Company, Device, 1);
  
  SMSDEV::DisplayWithDelay("Firmware", Version, 1);

	if(IsEEPROMValid(Signature)) {
		SMSDEV::DisplayWithDelay("EEPROM", "Valid", 1);
	} else {
		FactoryReset(NULL, 1);
	}
	ReadEEPromGlobals();
	
	SMSDEV::Display("Starting GSM...", "");
  if(GSMSMS::Init()) {
    SMSDEV::Display("Fatal Error!", "GSM Init Failed", 1);
  	return 1;
  }
  SMSDEV::DisplayWithDelay(NULL, "GSM Ready!");
  	
  SetParams();
  SendCmd("AT+CNMI=2,1");              // Enable Unsolicited SMS Notification
  SendCmd("AT+AUTOCSQ=1,1");           // Enable Unsolicited CSQ Notification (whenever it changes)
 
  strcpy(simcom, "SIM");
  itoa(GetSimCom(), simcom+3, 10);
  SMSDEV::DisplayWithDelay("GSM Chipset", simcom, 1);

	SMSDEV::Display("Checking SIM...", "");
	  if(! SIMInserted()) {
  	SMSDEV::Display("Fatal Error!", "No SIM Inserted", 1);
  	return 2;
  }
  SMSDEV::DisplayWithDelay(NULL, "SIM Detected!");

  DeleteAllSMS();
  
  GSMSMS::GetTime(TimeNow);
  DisplayTime();

  CSQ = GetSignalQuality();

  AuthCallers = DisplayAuthCallers(1);  // Just Count / Don't Display

  MessageTimer(MSGDISPLAYTIME);
  ScreenBlanker(SCREENBLANKTIME);
  
//  DisplayIdleScreen();			// Should be done by Calling Application at end of Setup();
  
  return 0;
}

int SMSDEV::Loop() {
static int tick = 0;
static int gsmcode = 0;
int smscode = 0;
int rcode = 0;

  if(tick == 0) tick = GetTime();
  if(tick) {
  	
  	if(CheckAlarm(Alarm1, EEALARM1)) return 101;			// Did Alarm1 just fire?
  	  	
  	if(CheckAlarm(Alarm2, EEALARM2)) return 102;      // Did Alarm2 just fire?
  	
		if(CheckHeartbeat()) return 103;                  // Did the Hearbeat just fire?
			
		if(PollSMS[0] == '1') {													  // Poll for SMS if enabled
			rcode = CheckSMS();
			if(rcode) return rcode;
		}

  	tick = 0;
  	return 100;
	}

  if(gsmcode == 0) gsmcode = CheckGSM();						  // Any GSM Messages Pending?
	switch (gsmcode) {
		case 1:																					  // SMS (CMTI) Notification
			smscode = CheckSMS();
			if(smscode) delay(1000);
			else	gsmcode = 0;															// (Handle multiple SMS Messages)
			return smscode;

		case 2:																					  // Signal (CSQ) Change
			CSQ = GetSignalQuality();
			gsmcode = 0;
			return 104;
	}

  CheckSerial();																			// Any Serial Commands Pending?

  MessageTimer();                           					// Does Display Need to be Updated?

  ScreenBlanker();                          					// Does the Display Need to be Blanked?
  
  return 0;
}

int SMSDEV::CheckGSM() {
int c;
int i;

  if(!Available()) return 0;

  Buffer[0]='\0';
  while(Available()) {
    c = Read(); 
    if ((i = strlen(Buffer)) < (BUFFSIZE - 1)) {
      Buffer[i] = c;  
      Buffer[i+1] = '\0';
    }
    if(!Available()) delay(100);
  }

  if(strstr(Buffer,"+CMTI:")) {
    return 1;
  }

  if(strstr(Buffer,"+CSQ:")) {
    return 2;
  }

  return 0;
}

int SMSDEV::CheckSMS() {
	
  if(GetSMSWithDisplay(Caller, Buffer) == 0) return 0;

	return ProcessSMS();
}

int SMSDEV::ProcessSMS() {
int IsCmd;
int CmdID;
int IsAuth;
int IsRestricted;
char *Buff;
  
  // Is this a Command?
  if(CmdPrefix != NULL) {
		if(strncasecmp(Buffer, CmdPrefix, strlen(CmdPrefix)) == 0) {
			Buff = Buffer + strlen(CmdPrefix);
			IsCmd = 1;
		} else {
			Buff = Buffer;
		  IsCmd = 0;
		}
	} else {
		Buff = Buffer;
		IsCmd = 1;
	}
	
	// Get the CommandID (GenCommands are Positive, SpecCommands will be made Negative)
	if(IsCmd) {
		CmdID = GetCommandID(Buff, GenCmds, NumGenCmds);
		if(CmdID == 0) {
			CmdID = GetCommandID(Buff, SpecCmds, NumSpecCmds);
			CmdID *= -1;
		}
	} else CmdID = 0;

  // Is caller authorized?
  if(AuthCallers == 0) {
    if(strlen(Caller) > 4) {
      if(AuthorizeCaller(Caller, 1)) IsAuth = 1;
    }
  } else IsAuth = CheckAuthorized(Caller);
  	
  // If Not Authorized, is this the PWD?
  if(!IsAuth) {
  	if(Pwd[0] != '\0' && strcasecmp(Buff, Pwd) == 0) {
      if(AuthorizeCaller(Caller, 1)) IsAuth=1;  
    }
  }
  
  // Is this a Restricted Command (Privs)
  IsRestricted = 0;
  if(CmdID < 0) {
  	if(SpecCmds[CmdID * -1].Restricted) IsRestricted = 1;
  } else {
  	if(GenCmds[CmdID].Restricted) IsRestricted = 1;  	
  }
  if(ProcessPrivs == 0) IsRestricted = 0;
	
	// Process (Gen) or Forward (Spec) the Command
  if(IsCmd) {
  	if(IsAuth) {
  		if(CmdID == 0) {
  			SendSMSWithDisplay(Caller, "What?");
  			return 201;														// Unknown Command
  		}
  		if(IsRestricted) {
  			ReadEEPROMField(EEAUTHCALLERS, Temp);
  			if(strcmp(Caller, Temp) != 0) {
  				SendSMSWithDisplay(Caller,"Only Primary Caller may issue this command!");
    			return 202;													// Insufficient Privs
  			}
  		}
  		if(CmdID < 0) return CmdID * -1;				// Device Specific Command (forward it)
  		return ProcessCommand(CmdID, Caller, Buff) * -1;    // Gen Command (process it)
  	} else {
  		ForwardMsg(Caller);
      return 200;  														// Unauthorized Caller!
  	}
  }

  return 203;																	// Not a Command, Forward for Processing
}

void SMSDEV::CheckSerial() {
int c;
char i;

  if(!Serial.available()) return;

  Buffer[0]='\0';
  while(Serial.available()) {
    c = Serial.read(); 
    if ((i = strlen(Buffer)) < (BUFFSIZE - 1)) {
      Buffer[i] = c;  
      Buffer[i+1] = '\0';
    }
    if(!Serial.available()) delay(100);
  }
  
  Serial.println();
  Serial.print("Cmd: ");
  Serial.print(Buffer);
  
  if(SendCmd(Buffer) == 0) Serial.println(GSMCommBuff);
  else {
    Serial.println();
    Serial.println("ERROR");
  }

}

int SMSDEV::ProcessCommand( int CmdID, char *Caller, char *Buff) {
char *p1;
	
	Param = strchr(Buff, ':');
	if(Param != NULL) Param = Param + strspn(Param, " :");

  switch (CmdID) {
  	case 1:   // PWD:
  		strcpy(Pwd, Param);
    	WriteEEPROMField(EEPWD, Param);
      SendSMSWithDisplay(Caller,"PWD Set!");       
      break;
      
  	case 2:   // DEVNUM:
  		strcpy(DevNum, Param);
    	WriteEEPROMField(EEDEVNUM, Param);
      SendCfgInfo(Caller);        
      break;
      
  	case 3:   // DEVNAME:      
  		strcpy(DevName, Param);
    	WriteEEPROMField(EEDEVNAME, Param);
  		SendCfgInfo(Caller);        
      break;
      
  	case 4:   // HBTIME:      
  		strcpy(HBTime, Param);
    	WriteEEPROMField(EEHBTIME, Param);
  		SendCfgInfo(Caller);        
      break;
      
  	case 5:   // BOOTNOTIFY:
  		if(!strcasecmp(Param, "on")) {
  			strcpy(BootNotify, "1");
  			WriteEEPROMField(EEBOOTNOTIFY, BootNotify);
  			SendCfgInfo(Caller); 
  		} else if(!strcasecmp(Param, "off")) {
  			strcpy(BootNotify, "0");
  			WriteEEPROMField(EEBOOTNOTIFY, BootNotify);
  			SendCfgInfo(Caller);
  		} else SendSMSWithDisplay(Caller,"What?");       
      break;
      
  	case 6:   // POLLSMS:
  		if(!strcasecmp(Param, "on")) {
  			strcpy(PollSMS, "1");
  			WriteEEPROMField(EEPOLLSMS, PollSMS);
	  		SendCfgInfo(Caller); 
  		} else if(!strcasecmp(Param, "off")) {
  			strcpy(PollSMS, "0");
  			WriteEEPROMField(EEPOLLSMS, PollSMS);
	  		SendCfgInfo(Caller);
  		} else SendSMSWithDisplay(Caller,"What?");       
      break;
      
  	case 7:   // CFGINFO
  		SendCfgInfo(Caller);      
      break;
      
  	case 8:   // AUTHLIST
  		SendAuthList(Caller);      
      break;
      
  	case 9:   // RESET 
  		FactoryReset(Caller);     
      break;
      
  	case 10:   // REBOOT
  		SendSMSWithDisplay(Caller,"Rebooting...");  
  		Reboot();        
      break;
      
  	case 11:   // SIMINFO
  		SendSIMInfo(Caller);   
      break;
      
  	case 12:   // HWINFO
  		SendHWInfo(Caller);      
      break;
      
  	case 13:   // CELLINFO  
  		SendCellInfo(Caller);    
      break;
      
  	case 14:   // OPERATORS
  		SendOperators(Caller);      
      break;
      
  	case 15:   // ATCMD:
  		p1=strchr(Param, '\n');
      if(p1 != NULL) *p1='\0';
      SendCmdParams(Param, 30000, 1000, "OK", 10);
      strncpy(Buffer, GSMCommBuff, BUFFSIZE);
      Buffer[BUFFSIZE-1] = '\0';
      SendSMSWithDisplay(Caller, Buffer);      
      break;
  }
  
  return CmdID;
}

int SMSDEV::GetCommandID(char *astring, CMDINFO *Commands, int NumCmds) {
int i;
 
  for (i=0; i<NumCmds; i++) {
    if(strncasecmp(astring, Commands[i].Cmd, strlen(Commands[i].Cmd)) == 0) return Commands[i].ID;
  }
  return 0;
}

int SMSDEV::SetPrivProcessing(int state) {
	ProcessPrivs = state;
	return ProcessPrivs;
}

int SMSDEV::CheckAuthorized(char *caller) {
int i;
char contents[20];

  for(i=0; i<EEMAXAUTH; i++) {
    ReadEEPROMField(EEAUTHCALLERS+i, contents);
    if(contents[0] == '\0') break;
    if(!strcmp(caller,contents)) return 1;  
  }
  
  return 0;
}

int SMSDEV::ForwardMsg(char *from) {
int i;
int pushval;
int msglen;
int oldloc;
int newloc;

  pushval = (6 + strlen(from) + 2);
  Buffer[BUFFSIZE-pushval-1]='\0';
  msglen = strlen(Buffer);
  for(i=msglen; i>0; i--) {
    oldloc=i-1;
    newloc=oldloc+pushval;
    Buffer[newloc] = Buffer[oldloc];
  }
  strncpy(Buffer,"(From ",6);
  strncpy(Buffer + 6, from, strlen(from));
  strncpy(Buffer + 6 + strlen(from), ") ", 2); 

  if(GetAuthCaller(0, Caller) == 0) return 0;
  	
  SendSMSWithDisplay(Caller, Buffer);

	return 1;
}

int SMSDEV::AuthorizeCaller(char *caller, int notify) {
char *authmsg="Authorized!";

  if(AuthCallers == EEMAXAUTH) {
    //  Send SMS stating max auth users?
    return 0;
  }
  
  WriteEEPROMField(EEAUTHCALLERS + AuthCallers, caller);
  AuthCallers++;
  
  if(notify) SendSMSWithDisplay(caller, authmsg);

  return AuthCallers;
}

void SMSDEV::FactoryReset(char *caller, int quiet) {
char auth0[EEFIELDSIZE];
	
	GetAuthCaller(0, auth0);

  Pwd[0]='\0';
	DevNum[0]='\0';
	DevName[0]='\0';
	strcpy(HBTime, "01,12:00");
	strcpy(BootNotify, "1");
	strcpy(PollSMS, "1");
	Alarm1[0]='\0';
	Alarm2[0]='\0';
	AuthCallers=0;

  InitializeEEPROM(Signature);
  DeleteAllSMS();
  
  SMSDEV::DisplayWithDelay("Factory", "Reset", 1);
  MessageTimer(-1);
  
  if(quiet) return;

  if(caller != NULL) SendSMSWithDisplay(caller, "Factory Reset! (Don't forget to re-authorize and set Password)");
  else if(strlen(auth0)) SendSMSWithDisplay(auth0, "Factory Reset!");	
}

// ALARM Functions ---------------------------------------------------------------------------------------------------------------- 

int SMSDEV::GetTime() {
unsigned long tnow;
static unsigned long ClockTimer = 0;

  tnow = millis();
  if (tnow < (ClockTimer + 30000)) return 0;     // Update the time every 30 secs to ensure we get 1 min resolution

  ClockTimer = tnow;
  GSMSMS::GetTime(TimeNow);
  return 1;
}

int SMSDEV::CheckHeartbeat() {
static int HBSent = 0;

  if(HBTime[0] == '/0') return 0;

  if(strncmp(TimeNow + 6, HBTime, strlen(HBTime)) == 0) {
    if(HBSent == 0) {
      HBSent = 1;
      return 1;
    }
    return 0;
  }
  
  HBSent = 0;
  return 0;
}

int SMSDEV::CheckAlarm(char *atime, int eetimefield) {
int i;
int j;
int tnow;
int alarm;
char tmp[3];
int expired = 0;

  if(strlen(atime) == 0) return 0;  // No Alarm set

  for(i=0; i<15; i+=3) {
    strncpy(tmp, TimeNow+i, 2);
    tnow=atoi(tmp);
    strncpy(tmp, atime+i, 2);
    alarm=atoi(tmp);
    
    if(alarm > tnow) return 0;  // Not yet...
    if(alarm < tnow) {
      expired = 1;		            // Expired
      break;
    }    
  }

  // Clear the Alarm in Local Memory
  atime[0]='\0';
  
  // Clear the Alarm in EEPROM
  EEPROM.write( eetimefield*EEFIELDSIZE, '\0');
#if defined(ESP32)
    EEPROM.commit();
#endif
  
  return expired ? 0 : 1;
}

// Funtions which return information --------------------------------------------------------

void SMSDEV::SendCfgInfo(char *acaller) {

  strcpy(Buffer,"Config Info:\n\n");
  
  strcat(Buffer,   "DevNum:     ");
  strcat(Buffer,    DevNum);
  
  strcat(Buffer, "\nDevName:    ");
  strcat(Buffer,    DevName);
  
  strcat(Buffer, "\nHBTime:     ");
  strcat(Buffer,    HBTime);
  
  strcat(Buffer, "\nBootNotify: ");
  if(BootNotify[0] == '1') strcat(Buffer, "On");
  else strcat(Buffer, "Off");
  	
  strcat(Buffer, "\nPollSMS:    ");
  if(PollSMS[0] == '1') strcat(Buffer, "On");
  else strcat(Buffer, "Off");

  SendSMSWithDisplay(acaller, Buffer);  
}

int SMSDEV::SendAuthList(char *acaller) {
int i;
int j = 0;
int maxauthsms = 5;
char msgend[2]={0x1a,'\0'};

  //  We are going to build our SMS Message incrementally using the low level library functions
  //  So we don't have to allocate a bunch of memory for a temp string
  for(i=0, j=0; i<EEMAXAUTH; i++) {

    if(GetAuthCaller(i, Temp)) {

      if(j == 0) {                            // Begin SMS Message
        if(PrepToSendSMS(acaller) > 0) return 1;
        if( i == 0) GSMSMS::Print("Authorized Callers List: ");
        else GSMSMS::Print("Authorized Callers List (cont): ");
        GSMSMS::Print("\n\n"); 
      }
    
      GSMSMS::Print("Auth ");
      GSMSMS::Print(i+1);
      GSMSMS::Print(": ");
      GSMSMS::Print(Temp);
      GSMSMS::Print("\n");
      j++; 
  
      if(j == maxauthsms - 1) {               // Send SMS Message
        IndicateSMSOut(acaller);
        GSMSMS::Print(msgend);
        if(FlushSMS() > 0) return 2;
        delay(500);
        j = 0;
      }
      
    } else break;
  }

  if(j) {                                   // Anything left to send?
    IndicateSMSOut(acaller);
    GSMSMS::Print(msgend);
    if(FlushSMS() > 0) return 2;
  }

  return 0;
}

void SMSDEV::SendSIMInfo(char *acaller) {
char *p1;
char *p2;
char *p3;

  strcpy(Buffer,"");

  // ICCID
  GetICCID(Temp);
  strcat(Buffer,"ICCID: ");
  strcat(Buffer,Temp);
  strcat(Buffer,"\n");
  
  // IMSI
  GetIMSI(Temp);
  strcat(Buffer,"IMSI: ");
  strcat(Buffer,Temp);
  strcat(Buffer,"\n");

  // Service Provider
  GetATResult("AT+CSPN?","+CSPN: ",Temp,1000);
  p1=strchr(Temp,'\"');
  if(p1 != NULL) {
    p1++;
    p2=strchr(p1,'\"');
    if(p2 != NULL) {
      *p2='\0';
      p3=p1;
    } else p3=Temp;
  } else p3=Temp;
  strcat(Buffer,"Provider: ");
  strcat(Buffer,p3);
  strcat(Buffer,"\n");
  
  // Subscriber Number
  GetSubscriberNumber(Temp);
  strcat(Buffer,"Number: ");
  strcat(Buffer,Temp);
  
  SendSMSWithDisplay(acaller,Buffer);
}

void SMSDEV::SendHWInfo(char *acaller) {
  
  strcpy(Buffer,"");

  // Chipset ID
  GetChipset(Temp);
  strcat(Buffer,"Chipset: ");
  strcat(Buffer,Temp);
  strcat(Buffer,"\n");
  
  // IMEI
  GetIMEI(Temp);
  strcat(Buffer,"IMEI: ");
  strcat(Buffer,Temp);

  SendSMSWithDisplay(acaller,Buffer);  
}

void SMSDEV::SendCellInfo(char *acaller) {
char *p1;
char *p2;
char *p3;

  strcpy(Buffer,"");

  // Current Band
  switch(GetSimCom()) {
    case 808:
    case 900:
      GetATResult("AT+CBAND?","+CBAND: ",Temp,1000);
      break;
    case 5320:
      GetATResult("AT*CNTI=0","*CNTI: 0,",Temp,1000);
      break;
  }
  strcat(Buffer,"Band: ");
  strcat(Buffer,Temp);
  strcat(Buffer,"\n");

  // Signal Quality
  sprintf(Temp,"Signal: %c/5",char(CSQ+48));
  strcat(Buffer,Temp);
  strcat(Buffer,"\n");
 
  // Current Network Operator and Mode
  GetATResult("AT+COPS?","+COPS: ",Temp,10000);
  p1=strchr(Temp,'\"');
  if(p1 != NULL) {
    p1++;
    p2=strchr(p1,'\"');
    if(p2 != NULL) {
      *p2='\0';
      p3=p1;
    } else p3=Temp;
  } else p3=Temp;
  strcat(Buffer,"Current: ");
  strcat(Buffer,p3);
  strcat(Buffer,"\n");
  
  // Service Center Address
  GetATResult("AT+CSCA?","+CSCA: ",Temp,1000);
  p1=strchr(Temp,'\"');
  if(p1 != NULL) {
    p1++;
    p2=strchr(p1,'\"');
    if(p2 != NULL) {
      *p2='\0';
      p3=p1;
    } else p3=Temp;
  } else p3=Temp;
  strcat(Buffer,"SMSC: ");
  strcat(Buffer,p3);
  strcat(Buffer,"\n");

  SendSMSWithDisplay(acaller,Buffer);  
}
 
void SMSDEV::SendOperators(char *acaller) {

  strcpy(Buffer,"");

  // Network Operators Currently in the Network
  GetATResult("AT+COPS=?","+COPS: ",Temp,60000);
  strcat(Buffer,"Operators: ");
  strcat(Buffer,Temp);
  strcat(Buffer,"\n");

  SendSMSWithDisplay(acaller,Buffer);  
}

// DISPLAY Functions -------------------------------------------------------------------------------------------------------------

int SMSDEV::GetSMSWithDisplay(char *number, char *msg) {
int stat;

  stat = GetSMS(number, msg, EEFIELDSIZE, BUFFSIZE);
  if(stat > 0) IndicateSMSIn(number);
  return stat;
}

int SMSDEV::SendSMSWithDisplay(char *number, char *msg) {
  IndicateSMSOut(number);
  return SendSMS(number, msg);
}

void SMSDEV::DisplayIdleScreen() {
char ALine[20];

  if(CSQ == 0) {
    SMSDEV::Display("Poor Signal","No Service",1);
    return;
  }
  
  if(AuthCallers == 0) {
    sprintf(ALine,"Call!        %c/5",char(CSQ + 48));
    SMSDEV::Display("Waiting for 1st", ALine);
    return;
  }

  sprintf(ALine,"%-12s %c/5",Device, char(CSQ + 48));
  if(strlen(DevNum)) SMSDEV::Display(ALine, DevNum, 1);
  else SMSDEV::Display("Haute Solutions", ALine); 
}

void SMSDEV::DisplayTime() {
  
  SMSDEV::DisplayWithDelay("Time Now",TimeNow, 1);
}

int SMSDEV::DisplayAuthCallers(int JustCount) {
int i;

  for(i=0; i<EEMAXAUTH; i++) {
    if(GetAuthCaller(i, Caller)) {
      if(!JustCount) SMSDEV::DisplayWithDelay("Auth Caller:", Caller);
    } else {
    	if(i == 0 && !JustCount) SMSDEV::DisplayWithDelay("Auth Callers", "(None)", 1);
    	return i;   
    }
  }
  
  return i+1;
}

void SMSDEV::DisplayCfgInfo() {
	
	SMSDEV::DisplayWithDelay("Device Name", DevName, 1);

	SMSDEV::DisplayWithDelay("Device Number", DevNum, 1);

	SMSDEV::DisplayWithDelay("HeartBeat Time", HBTime, 1);

	if(BootNotify[0] == '1') SMSDEV::DisplayWithDelay("Boot Notify", "On", 1);
	else SMSDEV::DisplayWithDelay("Boot Notify", "Off", 1);

  if(PollSMS[0] == '1') SMSDEV::DisplayWithDelay("Poll SMS", "On", 1);
  else SMSDEV::DisplayWithDelay("Poll SMS", "Off", 1);
 
}

void SMSDEV::IndicateSMSIn(char *Number) {
  SMSDEV::DisplayWithDelay("SMS IN from:", Number);
  MessageTimer(-1);
}

void SMSDEV::IndicateSMSOut(char *Number) {
  SMSDEV::DisplayWithDelay("SMS OUT to:", Number);
  MessageTimer(-1);
}

void SMSDEV::RollDisplay(char *Line2, int Center) {
	
	ScreenState(1);
  if(OLEDAddr != NULL) OLED::RollDisplay(Line2, Center);
}

void SMSDEV::Display(char *Line1, char *Line2, int Center) {
	
	ScreenState(1);
  if(OLEDAddr != NULL) OLED::Display(Line1, Line2, Center);
}

void SMSDEV::DisplayWithDelay(char *Line1, char *Line2, int Center) {
	
	ScreenState(1);
  if(OLEDAddr != NULL) OLED::Display(Line1, Line2, Center);
	delay(MSGDISPLAYTIME);
}

int SMSDEV::ScreenState(int state) {
static int ScreenState = 0;

	switch (state) {
		case 0:
			if(ScreenState == 1) {
				OLED::Off();
				ScreenState = 0;
			}
			break;
		case 1:
			if(ScreenState == 0) {
				OLED::On();
				ScreenState = 1;
			}
			ScreenBlanker(-1);
			break;
	}
	
	return ScreenState;
}

void SMSDEV::ScreenBlanker(int mode) {
static unsigned long ScreenTimer;
static unsigned long ScreenBlankTime;	
	
	if(mode < 0) {						// Reset the Timer
		ScreenTimer = millis();
	} else if (mode > 0) {		// Set The Duration
		ScreenBlankTime = mode;	
	} else {									// Check the Timer
		if(ScreenState() == 0) return;
		if(ScreenBlankTime == 0) return;
    if(millis() < (ScreenTimer + ScreenBlankTime)) return;
    ScreenState(0);
	}

}

void SMSDEV::MessageTimer(int mode) {
static unsigned long MsgTimer;
static unsigned long MsgDisplayTime;

	if(mode < 0) {						// Reset the Timer
		MsgTimer = millis();
	} else if (mode > 0) {		// Set The Duration
		MsgDisplayTime = mode;	
	} else {									// Check the Timer
		if(MsgTimer == 0) return;
		if(MsgDisplayTime == 0) return;
    if(millis() < (MsgTimer + MsgDisplayTime)) return;
    DisplayIdleScreen();
    MsgTimer = 0;
	}
}

// ----------------------------- Routines which deal with Persistent (EEPROM Data) ---------------------------------------------

int SMSDEV::IsEEPROMValid(char *sig) {

  if(ReadEEPROMField(EESIG, Buffer) < strlen(sig)) return 0;
  if(strcmp(Buffer, sig)) return 0;
  return 1;
}

void SMSDEV::InitializeEEPROM(char *sig) {
int i;

  for(i = 0; i < EEAUTHCALLERS + EEMAXAUTH; i++) EEPROM.write(i * EEFIELDSIZE, '\0');  
  	
  WriteEEPROMField(EESIG, sig);

  WriteEEPROMGlobals();
}

int SMSDEV::ReadEEPromGlobals() {
	
	ReadEEPROMField(EEPWD,        Pwd);
	ReadEEPROMField(EEDEVNUM,     DevNum);
	ReadEEPROMField(EEDEVNAME,    DevName);
	ReadEEPROMField(EEHBTIME,     HBTime);
	ReadEEPROMField(EEBOOTNOTIFY, BootNotify);
	ReadEEPROMField(EEPOLLSMS,    PollSMS);
	ReadEEPROMField(EEALARM1,     Alarm1);
	ReadEEPROMField(EEALARM2,     Alarm2);
}

int SMSDEV::WriteEEPROMGlobals() {
	
	WriteEEPROMField(EEPWD,        Pwd);
	WriteEEPROMField(EEDEVNUM,     DevNum);
	WriteEEPROMField(EEDEVNAME,    DevName);
	WriteEEPROMField(EEHBTIME,     HBTime);
	WriteEEPROMField(EEBOOTNOTIFY, BootNotify);
	WriteEEPROMField(EEPOLLSMS,    PollSMS);
	WriteEEPROMField(EEALARM1,     Alarm1);
	WriteEEPROMField(EEALARM2,     Alarm2);
}

int SMSDEV::GetAuthCaller(int pos, char *caller) {
  if (ReadEEPROMField(EEAUTHCALLERS + pos, caller) == 0) return 0;
  return strlen(caller);
}

void SMSDEV::SetAuthCaller(int pos, char *caller) {
  WriteEEPROMField(EEAUTHCALLERS + pos, caller);
}

int SMSDEV::WriteEEPROMField(int field, char *text) {
  return WriteEEPROMString( field*EEFIELDSIZE, text, EEFIELDSIZE);
}

int SMSDEV::ReadEEPROMField(int field, char *text) {
  return ReadEEPROMString( field*EEFIELDSIZE, text, EEFIELDSIZE);
}

int SMSDEV::WriteEEPROMString(int offset, char *text, int maxlen) {
int i;
int c;
int status;

  if(text == NULL || !strlen(text)) {
  	EEPROM.write(offset, '\0');
  	status = 0;
  } else {
		for (i=0; i<maxlen; i++) {
		  if(i == (maxlen-1)) EEPROM.write(offset+i, '\0');
		  else EEPROM.write(offset+i, text[i]);
		  if(text[i] == '\0') break;
		}
		status = i+1;
	}
  
#if defined(ESP32)
    EEPROM.commit();
#endif

  return status;
}

int SMSDEV::ReadEEPROMString(int offset, char *text, int maxlen) {
int i;

  for (i=0; i<maxlen; i++) {
    text[i] = EEPROM.read(offset+i);
    if(text[i] == '\0') break;
  }
  
  text[maxlen-1]='\0';

  return i+1;
}

