// GSMSMS Library
// Copyright (c) 2020 by Haute Solutions / Ed Van Every (www.haute-solutions.com)
//
// FUNCTION: This Class Library is intended to simplify command processing for SMS Enabled Devices.
//           It inherits objects and methods from the Haute Solutions GSMSMS and Haute Solutions OLED
//           class libraries.  OLED Functionality is optional (Init with NULL)
//
// HARDWARE: Please refer to the GSMSMS and OLED Class Libraries for Hardware Support Information
//
// SOFTWARE: Requires the Haute Solutions GSMSMS and Haute Solutions OLED Class Libraries
//
// LICENSE:  This library is free software: you can redistribute it and/or modify
//           it under the terms of the GNU General Public License as published by
//           the Free Software Foundation, either version 3 of the License, or
//           (at your option) any later version.
//
//           This program is distributed in the hope that it will be useful,
//           but WITHOUT ANY WARRANTY; without even the implied warranty of
//           MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//           GNU General Public License for more details.
//           <http://www.gnu.org/licenses/>.
//
// VERSION:  SMSDEV_LIBVER "1.3"

#ifndef SMSDEV_h
#define SMSDEV_h

#define SMSDEV_LIBVER "1.3"

#include "Arduino.h"
#include "GSMSMS.h"
#include "EEPROM.h"
#include "OLED_SSD1311.h"

// Persistant Storage Info (EEPROM Info): Character Data, NULL Terminated
// All Fields are 20 Chars max (Inc NULL Terminator)
// The values below are the Field Offsets into EEPROM memory for start of each string
// Currently (V3 EEPROM Layout) Uses 800 Bytes Total
// (Lib Generic Data + Dev Specific Data + Auth Data = 10*20 + 10*20 + 20*20

#define EENUMFIELDS     40               // Total Number of EEPROM Fields
#define EEFIELDSIZE     20               // Size of each field/string (inc Terminator)
#define EEMAXAUTH       20               // Max Number of Authorized Callers

#define EESIG            0               // EEPROM Data Valid Signature
#define EEPWD            1               // Password for Self Authorization
#define EEDEVNUM         2               // Device Cell Number
#define EEDEVNAME        3               // Device Name
#define EEHBTIME         4               // Heartbeat Time
#define EEBOOTNOTIFY     5               // Notify on Boot ("1" | "0")
#define EEPOLLSMS        6               // Poll SMS (at each tick)  ("1" | "0")
#define EEALARM1         7               // Alarm 1
#define EEALARM2         8               // Alarm 2
#define EEAVAIL          9               // {Future Expansion}

#define EEDEVSPEC1      10               // Device Specific Data Storage
#define EEDEVSPEC2      11               // Device Specific Data Storage
#define EEDEVSPEC3      12               // Device Specific Data Storage
#define EEDEVSPEC4      13               // Device Specific Data Storage
#define EEDEVSPEC5      14               // Device Specific Data Storage
#define EEDEVSPEC6      15               // Device Specific Data Storage
#define EEDEVSPEC7      16               // Device Specific Data Storage
#define EEDEVSPEC8      17               // Device Specific Data Storage
#define EEDEVSPEC9      18               // Device Specific Data Storage
#define EEDEVSPEC10     19               // Device Specific Data Storage

#define EEAUTHCALLERS   20               // Start of Authorized Caller Phone Numbers Array

#define BUFFSIZE 300                     // Size of Work Buffer
#define TIMESIZE 15

#define SCREENBLANKTIME 30000            // Time to wait before turning off display
#define MSGDISPLAYTIME  2000             // Time to display info messages (before going back to idle screen)

typedef struct CMDINFO{
  char* Cmd;
  int ID;
  bool Restricted;
};


class SMSDEV : public GSMSMS, public OLED
{
	public:
		
		char Sig[EEFIELDSIZE];
		char Pwd[EEFIELDSIZE];
		char DevNum[EEFIELDSIZE];
		char DevName[EEFIELDSIZE];
		char HBTime[EEFIELDSIZE];
		char BootNotify[EEFIELDSIZE];
		char PollSMS[EEFIELDSIZE];
		char Alarm1[EEFIELDSIZE];
		char Alarm2[EEFIELDSIZE];
	
		char TimeNow[TIMESIZE];
		int  CSQ;
		int  AuthCallers = 0;
		char Caller[EEFIELDSIZE];
		char Buffer[BUFFSIZE];
		char Temp[BUFFSIZE];
		char *Param;

		SMSDEV(int board, int oledaddr) : GSMSMS(board), OLED(oledaddr) {
			Board = board;
			OLEDAddr = oledaddr;
		}
		
		int   Setup(char *company, char* device, char* version, char* sig, char* cmdprefix, CMDINFO* cmds, int numcmds);
		int   Loop();
		int   CheckGSM();
		void  CheckSerial();
		int   CheckSMS();
		int   ProcessSMS();
		int   ProcessCommand(int CmdID, char *caller, char *Buff);
		int   GetCommandID(char *astring, CMDINFO *Commands, int NumCmds);
		int   SetPrivProcessing(int state);
		int   CheckAuthorized(char *caller);
		int   ForwardMsg(char *from);
		int   AuthorizeCaller(char *caller, int notify=0);
		void  FactoryReset(char *caller = NULL, int quiet = 0);
		void  (* Reboot) (void) = 0;                     //declare reset function @ address 0

		int   GetTime();
		int   CheckHeartbeat();
		int   CheckAlarm(char *atime, int eetimefield);

		void  SendCfgInfo(char *acaller);
		int   SendAuthList(char *acaller);
		void  SendSIMInfo(char *acaller);
		void  SendHWInfo(char *acaller);
		void  SendCellInfo(char *acaller);
		void  SendOperators(char *acaller);

		int   GetSMSWithDisplay(char *number, char *msg);
		int   SendSMSWithDisplay(char *number, char *msg);
		void  DisplayIdleScreen();
		void  DisplayTime();
		int   DisplayAuthCallers(int JustCount=0);
		void  DisplayCfgInfo();
		void  IndicateSMSIn(char *Number);
		void  IndicateSMSOut(char *Number);
		void  RollDisplay(char *Line2, int Center=0);
		void  Display(char *Line1, char *Line2, int Center=0);
		void  DisplayWithDelay(char *Line1, char *Line2, int Center=0);
		int   ScreenState(int state=3);
		void  ScreenBlanker(int mode=0);
		void  MessageTimer(int mode=0);

		int   IsEEPROMValid(char *sig);
		void  InitializeEEPROM(char *sig);
		int   ReadEEPromGlobals();
		int   WriteEEPROMGlobals();
		int   GetAuthCaller(int pos, char *caller);
		void  SetAuthCaller(int pos, char *caller);
		int   WriteEEPROMField(int field, char *text);
		int   ReadEEPROMField(int field, char *text);
		int   WriteEEPROMString(int offset, char *text, int maxlen);
		int   ReadEEPROMString(int offset, char *text, int maxlen);

	private:
		int      Board;
		int      OLEDAddr;
		char*    Company;
		char*    Device;
		char*    Version;
		char*    Signature;
		char*    CmdPrefix;
		int      NumGenCmds;
		CMDINFO* SpecCmds;
		int      NumSpecCmds;
		int      ProcessPrivs;
};

#endif





